"""
Test for decorated function.
"""
from src import decorated


def test_decorated():
    """
    Test that decremented square of 3 is 8.
    """
    assert decorated.decremented_square(3) == 8


def test_decorated_2():
    """
    Test that 2 * 5 + 3 = 13
    """
    assert decorated.modified_double(5) == 13



"""Code for pytest-rts plugin logic"""
import logging
import os

import pytest
from _pytest.config import Config
from _pytest.config.argparsing import Parser

from pytest_rts.pytest.runner_plugin import RunnerPlugin
from pytest_rts.utils.common import (
    get_existing_tests,
    get_tests_from_changes,
)
from pytest_rts.utils.git import is_git_repo


def pytest_addoption(parser: Parser) -> None:
    """Register pytest flags"""
    group = parser.getgroup("pytest-rts")
    group.addoption("--rts", action="store_true", default=False, help="Run pytest-rts")
    group.addoption(
        "--rts-coverage-db",
        action="store",
        default="",
        help="Coverage file for pytest-rts",
    )
    group.addoption(
        "--rts-from-commit",
        action="store",
        default="",
        help="Give a comparison Git commithash for RTS",
    )


def pytest_configure(config: Config) -> None:
    """Register RTS plugins based on state"""
    if not config.option.rts:
        return

    if not is_git_repo():
        pytest.exit("pytest-rts: Cannot find a Git repository in the current folder", 2)

    if not config.option.rts_coverage_db:
        pytest.exit("pytest-rts: No coverage file provided", 2)

    if not os.path.exists(config.option.rts_coverage_db):
        pytest.exit("pytest-rts: Provided coverage file does not exist", 2)

    logging.basicConfig(format="%(message)s", level=logging.INFO)

    existing_tests = get_existing_tests(config.option.rts_coverage_db)
    tests_from_changes = get_tests_from_changes(
        config.option.rts_from_commit, config.option.rts_coverage_db
    )
    config.pluginmanager.register(RunnerPlugin(existing_tests, tests_from_changes))



"""Unittests for Git functionality"""
from pytest_rts.utils.git import get_changed_lines


def test_get_changed_lines() -> None:
    """Test case for reading changed lines from git diff -U0 output"""
    diff = """diff --git a/pytest_rts/utils/common.py b/pytest_rts/utils/common.py
            index 43d1651..63be8a8 100644
            --- a/pytest_rts/utils/common.py
            +++ b/pytest_rts/utils/common.py
            @@ -18 +18 @@ def filter_pytest_items(
            -    return list(
            +    return list(set(
            @@ -21,0 +22,2 @@ def filter_pytest_items(
            +
            +                new_var = 1
            @@ -42,2 +43,0 @@ def get_existing_tests(coverage_file_path: str) -> Set[str]:
            -
            -
            @@ -86 +85,0 @@ def strip_pytest_cov_testname(testname: str) -> str:
            -    return testname
    """
    changed_lines = get_changed_lines(diff)
    assert changed_lines == {18, 21, 42, 43, 86}



"""Integration tests for pytest-rts"""
import os
import subprocess
from typing import Dict, List

import pytest
from _pytest.pytester import Testdir

from pytest_rts.utils.git import get_git_repo

COV_FILE = "rts-coverage.db"


def init_git_repo(tmp_testdir_path: str) -> None:
    """Helper function to initialize a Git repository during tests"""
    if os.getcwd() != tmp_testdir_path:
        return
    subprocess.run(["git", "init"], check=True)
    subprocess.run(["git", "config", "user.name", "pytest"], check=True)
    subprocess.run(["git", "config", "user.email", "pytest@example.com"], check=True)
    subprocess.run(["git", "add", "."], check=True)
    subprocess.run(["git", "commit", "-m", "1"], check=True)


def commit_change(tmp_testdir_path: str, file_path: str) -> None:
    """Helper function to commit a change to the temporary repo during tests"""
    if os.getcwd() != tmp_testdir_path:
        return
    subprocess.run(
        ["git", "add", file_path],
        check=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    subprocess.run(
        ["git", "commit", "-m", f"commit {file_path}"],
        check=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


def change_file(
    tmp_testdir_path: str, path_of_change: str, path_to_change: str
) -> None:
    """Helper function to change a file in the testdir during tests"""
    if os.getcwd() != tmp_testdir_path:
        return
    subprocess.run(
        ["cp", "-f", path_of_change, path_to_change],
        check=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


def cleanup_caches(tmp_testdir_path: str) -> None:
    """Helper function to cleanup python and pytest caches
    in testdir between test runs to prevent flakiness
    """
    if os.getcwd() != tmp_testdir_path:
        return
    subprocess.run(
        ["rm", "-rf", "src/__pycache__/"],
        check=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    subprocess.run(
        ["rm", "-rf", "tests/__pycache__/"],
        check=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    subprocess.run(
        ["rm", "-rf", ".pytest_cache/"],
        check=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


def get_testrepo_commithash(tmp_testdir_path: str) -> str:
    """Helper function to return the current Git HEAD in the testrepo"""
    if os.getcwd() != tmp_testdir_path:
        return ""
    repo = get_git_repo()
    return repo.repo.head.object.hexsha


def squash_commits(tmp_testdir_path: str, num_to_squash: int) -> None:
    """Helper function to squash given amount of commits in the testrepo"""
    if os.getcwd() != tmp_testdir_path:
        return
    subprocess.run(["git", "reset", "--soft", f"HEAD~{num_to_squash}"], check=True)
    subprocess.run(["git", "commit", "-m", "squashed_commit"], check=True)


@pytest.fixture(name="testrepo")
def fixture_testrepo(testdir: Testdir) -> Testdir:
    """Run setup for tests in the testdir"""
    testdir.copy_example(".")
    init_git_repo(str(testdir.tmpdir))
    testdir.runpytest_subprocess("--cov=.", "--cov-context=test")
    os.rename(".coverage", COV_FILE)
    cleanup_caches(str(testdir.tmpdir))
    return testdir


def test_only_new_functions_are_ran(testrepo: Testdir) -> None:
    """Test case for running pytest-rts when new tests are added
    and changes are not committed.
    """

    # add a new test method to an existing file
    change_file(
        str(testrepo.tmpdir),
        "changes/test_car/add_test_passengers.txt",
        "tests/test_car.py",
    )

    # add a new test file with a test in it
    testrepo.makepyfile(
        """
        def test_empty():
            assert 1 == 1
    """
    )

    # run pytest-rts to only run new tests
    result = testrepo.runpytest("--rts", f"--rts-coverage-db={COV_FILE}")
    result.assert_outcomes(passed=2, failed=0, errors=0)


@pytest.mark.parametrize(
    "path_of_change, path_to_change, result_dict",
    [
        (
            "changes/shop/change_get_price.txt",
            "src/shop.py",
            {"passed": 2, "failed": 0, "errors": 0},
        ),
        (
            "changes/decorated/change_decorator.txt",
            "src/decorators.py",
            {"passed": 0, "failed": 1, "errors": 0},
        ),
        (
            "changes/decorated/change_decorator_2.txt",
            "src/decorators.py",
            {"passed": 0, "failed": 1, "errors": 0},
        ),
        (
            "changes/init/change_function_one.txt",
            "src/__init__.py",
            {"passed": 0, "failed": 1, "errors": 0},
        ),
    ],
)
def test_changes_in_working_directory(
    path_of_change: str,
    path_to_change: str,
    result_dict: Dict[str, int],
    testrepo: Testdir,
) -> None:
    """Test cases for running pytest-rts when changes
    are made in the Git working directory
    """
    # do a change
    change_file(str(testrepo.tmpdir), path_of_change, path_to_change)

    # run pytest-rts to only run tests based on changes
    result = testrepo.runpytest("--rts", f"--rts-coverage-db={COV_FILE}")
    result.assert_outcomes(
        passed=result_dict["passed"],
        failed=result_dict["failed"],
        errors=result_dict["errors"],
    )


def test_usage_no_git(testdir: Testdir) -> None:
    """Test case for pytest-rts usage in non-git directory"""
    testdir.copy_example(".")
    testdir.runpytest_subprocess("--cov=.", "--cov-context=test")
    os.rename(".coverage", COV_FILE)
    result = testdir.runpytest_subprocess("--rts", f"--rts-coverage-db={COV_FILE}")
    assert result.ret == 2


@pytest.mark.parametrize(
    "flags, returncode",
    [
        (["--rts"], 2),
        (["--rts", "--rts-coverage-db=does-not-exist"], 2),
    ],
)
def test_misused_flags(flags: List[str], returncode: int, testrepo: Testdir) -> None:
    """Test cases for misuse of pytest-rts flags that cause pytest to exit"""
    result = testrepo.runpytest_subprocess(*flags)
    assert result.ret == returncode


@pytest.mark.parametrize(
    "path_of_change, path_to_change, result_dict",
    [
        (
            "changes/shop/change_get_price.txt",
            "src/shop.py",
            {"passed": 2, "failed": 0, "errors": 0},
        ),
        (
            "changes/decorated/change_decorator.txt",
            "src/decorators.py",
            {"passed": 0, "failed": 1, "errors": 0},
        ),
        (
            "changes/decorated/change_decorator_2.txt",
            "src/decorators.py",
            {"passed": 0, "failed": 1, "errors": 0},
        ),
        (
            "changes/init/change_function_one.txt",
            "src/__init__.py",
            {"passed": 0, "failed": 1, "errors": 0},
        ),
    ],
)
def test_committed_changes(
    path_of_change: str,
    path_to_change: str,
    result_dict: Dict[str, int],
    testrepo: Testdir,
) -> None:
    """Test cases for running pytest-rts when changes are committed"""
    original_commithash = get_testrepo_commithash(str(testrepo.tmpdir))
    change_file(str(testrepo.tmpdir), path_of_change, path_to_change)
    commit_change(str(testrepo.tmpdir), path_to_change)

    result = testrepo.runpytest(
        "--rts",
        f"--rts-coverage-db={COV_FILE}",
        f"--rts-from-commit={original_commithash}",
    )
    result.assert_outcomes(
        passed=result_dict["passed"],
        failed=result_dict["failed"],
        errors=result_dict["errors"],
    )


def test_commithash_does_not_exist(testrepo: Testdir) -> None:
    """Test case for invalid given commithash where only
    tests from working directory should be collected
    """
    change_file(
        str(testrepo.tmpdir),
        "changes/decorated/change_decorator.txt",
        "src/decorators.py",
    )
    commit_change(str(testrepo.tmpdir), "src/decorators.py")

    change_file(
        str(testrepo.tmpdir),
        "changes/shop/change_get_price.txt",
        "src/shop.py",
    )

    result = testrepo.runpytest(
        "--rts",
        f"--rts-coverage-db={COV_FILE}",
        "--rts-from-commit=does-not-exist",
    )

    result.assert_outcomes(
        passed=2,
        failed=0,
        errors=0,
    )


def test_multiple_commits(testrepo: Testdir) -> None:
    """Test case for running pytest-rts after two commits"""
    original_commithash = get_testrepo_commithash(str(testrepo.tmpdir))
    change_file(
        str(testrepo.tmpdir), "changes/shop/change_get_price.txt", "src/shop.py"
    )
    commit_change(str(testrepo.tmpdir), "src/shop.py")

    change_file(
        str(testrepo.tmpdir),
        "changes/test_car/add_test_passengers.txt",
        "tests/test_car.py",
    )
    commit_change(str(testrepo.tmpdir), "tests/test_car.py")

    result = testrepo.runpytest(
        "--rts",
        f"--rts-coverage-db={COV_FILE}",
        f"--rts-from-commit={original_commithash}",
    )
    result.assert_outcomes(
        passed=3,
        failed=0,
        errors=0,
    )


def test_squashing_commits(testrepo: Testdir) -> None:
    """Test case for running pytest-rts after doing two commits
    and then squashing them into one
    """
    original_commithash = get_testrepo_commithash(str(testrepo.tmpdir))

    change_file(
        str(testrepo.tmpdir), "changes/shop/change_get_price.txt", "src/shop.py"
    )
    commit_change(str(testrepo.tmpdir), "src/shop.py")

    change_file(
        str(testrepo.tmpdir),
        "changes/test_car/add_test_passengers.txt",
        "tests/test_car.py",
    )
    commit_change(str(testrepo.tmpdir), "tests/test_car.py")
    squash_commits(str(testrepo.tmpdir), 2)

    result = testrepo.runpytest(
        "--rts",
        f"--rts-coverage-db={COV_FILE}",
        f"--rts-from-commit={original_commithash}",
    )
    result.assert_outcomes(
        passed=3,
        failed=0,
        errors=0,
    )


def test_both_committed_and_workdir(testrepo: Testdir) -> None:
    """Test case for running tests when there are changes
    both committed and in the working directory
    """
    original_commithash = get_testrepo_commithash(str(testrepo.tmpdir))

    change_file(
        str(testrepo.tmpdir),
        "changes/decorated/change_decorator.txt",
        "src/decorators.py",
    )
    commit_change(str(testrepo.tmpdir), "src/decorators.py")

    change_file(
        str(testrepo.tmpdir),
        "changes/shop/change_get_price.txt",
        "src/shop.py",
    )

    result = testrepo.runpytest(
        "--rts",
        f"--rts-coverage-db={COV_FILE}",
        f"--rts-from-commit={original_commithash}",
    )

    result.assert_outcomes(
        passed=2,
        failed=1,
        errors=0,
    )



"""
Decorator, to check that changes in the decorator switch on tests using
decorated functions.
"""
from functools import wraps
from typing import Callable

def decrement(func: Callable[[int], int]) -> Callable[[int], int]:
    """
    Decrements the value returned by the decorated function.
    """
    @wraps(func)
    def _wrapper(num):
        return func(num) - 1

    return _wrapper


def modify_by(delta: int):
    """
    Adds delta to the value returned by the decorated function.
    """
    def decorator(func: Callable[[int], int]) -> Callable[[int], int]:
        @wraps(func)
        def _wrapper(num):
            return func(num) + delta
        return _wrapper
    return decorator




"""This module contains a fake pytest item class"""
from _pytest.config import Config


class FakeItem:  # pylint: disable=too-few-public-methods
    """Fake class"""

    def __init__(self, config: Config) -> None:
        self.config = config



"""
Decorated function, to test that changes in the decorator code switch on tests
using the function itself.
"""

from .decorators import decrement, modify_by

@decrement
def decremented_square(num: int) -> int:
    """Square the number and decrement the result."""
    return num * num


@modify_by(3)
def modified_double(num: int) -> int:
    """Double the number and and three to the result."""
    return 2 * num



"""This module contains fuctions for test selecting operations"""
from typing import List, Set

from coverage import CoverageData
from _pytest.nodes import Item

from pytest_rts.utils.git import (
    commit_exists,
    get_git_repo,
    get_changed_lines,
    get_changed_files_workdir,
    get_changed_files_committed_and_workdir,
    get_file_diff_data_workdir,
    get_file_diff_data_committed_and_workdir,
)


def filter_pytest_items(
    pytest_items: List[Item], existing_tests: Set[str], tests_from_changes: Set[str]
) -> List[Item]:
    """Select pytest items if they are new and not marked as skipped"""
    return list(
        filter(
            lambda item: (item.nodeid in tests_from_changes)
            or (
                item.nodeid not in existing_tests
                and not item.get_closest_marker("skipif")
                and not item.get_closest_marker("skip")
            ),
            pytest_items,
        )
    )


def get_existing_tests(coverage_file_path: str) -> Set[str]:
    """Read all the test function names from the coverage file.
    pytest-cov creates the coverage file and adds a section at the
    end of each testname which need to be stripped.
    """
    coverage_data = CoverageData(coverage_file_path)
    coverage_data.read()
    return {
        strip_pytest_cov_testname(testname)
        for testname in coverage_data.measured_contexts()
    }


def get_tests_from_changes(
    commithash_to_compare: str, coverage_file_path: str
) -> Set[str]:
    """Returns the test set from Git changes.
    The given commithash is compared to the current working copy
    to extract Git diffs, if the provided commithash exists in the repo.
    Otherwise only changes in the git working directory are considered.
    """
    repo = get_git_repo()
    if commit_exists(commithash_to_compare, repo):
        file_diffs = {
            file_path: get_file_diff_data_committed_and_workdir(
                repo, file_path, commithash_to_compare
            )
            for file_path in get_changed_files_committed_and_workdir(
                repo, commithash_to_compare
            )
        }
    else:
        file_diffs = {
            file_path: get_file_diff_data_workdir(repo, file_path)
            for file_path in get_changed_files_workdir(repo)
        }
    coverage_data = CoverageData(coverage_file_path)
    coverage_data.read()

    tests: Set[str] = set()
    for changed_file in file_diffs:

        contexts = coverage_data.contexts_by_lineno(changed_file)
        if not contexts:
            continue

        changed_lines_with_tests = intersect_with_surroundings(
            get_changed_lines(file_diffs[changed_file]),
            contexts.keys()
        )

        tests.update(
            strip_pytest_cov_testname(testname)
            for line in changed_lines_with_tests
            for testname in contexts[line]
        )

    return tests


def intersect_with_surroundings(changed_lines: Set[int], mapped_lines: Set[int]) -> Set[int]:
    """
    Finds lines existing in both changed_lines and mapped_lines.
    For lines that exist in changed_lines but not in mapped_lines
    looks for the closest mapped lines from left and right
    hand sides - mapped lines which surround changed line.

    Example:
        changed_lines | 1       5             21    30
        mapped_lines  |    2 3    10 11 12 20 21 22
        result        |    2 3    10          21 22

    Algorithm could be improved. See conversation:
    https://github.com/F-Secure/pytest-rts/pull/103#pullrequestreview-625312058
    """
    mapped = changed_lines.intersection(mapped_lines)
    unmapped = changed_lines.difference(mapped)

    mapped_lines_sorted = sorted(mapped_lines)
    for line in unmapped:
        left = None
        right = None
        for mapped_line in mapped_lines_sorted:
            if mapped_line < line:
                left = mapped_line
            if mapped_line > line:
                right = mapped_line
                break
        if left is not None:
            mapped.add(left)
        if right is not None:
            mapped.add(right)
    return mapped


def strip_pytest_cov_testname(testname: str) -> str:
    """Strip ends of pytest-cov produced testnames"""
    if testname.endswith("|teardown"):
        return testname[:-9]
    if testname.endswith("|setup"):
        return testname[:-6]
    if testname.endswith("|run"):
        return testname[:-4]
    return testname




class Shop:

    def __init__(self,item_price,items,money):
        self.item_price = item_price
        self.items = items
        self.money = money

    def buy_item(self):
        if self.items > 0:
            self.items = self.items - 1
            self.money = self.money + self.item_price
        
    def get_items(self):
        return self.items

    def get_item_price(self):
        return self.item_price

    def get_money(self):
        return self.money

    def calculate_useless_stuff(self):
        k = 2
        s = "ATGATATCATCGACGATGTAG"
        context = {}  
        for i,c in enumerate(s):
            if (i+k >= len(s)):break
            substring = s[i:i+k]
            if substring not in context:
                context[substring] = s[i+k]
            else:
                context[substring] = context[substring] + s[i+k]            
    
        return context


    



"""This module contains code for git operations"""
import logging
import re
import subprocess
from typing import List, Set

from gitdb.exc import BadName
from pydriller import Git


def commit_exists(commithash: str, repo: Git) -> bool:
    """Check if a given commithash exists in the Git repository"""
    if not commithash:
        return False
    try:
        repo.get_commit(commithash)
        return True
    except BadName:
        logging.info("Git commithash was provided but it was not found in history.")
        return False


def get_changed_files_workdir(repo: Git) -> List[str]:
    """Get changed files in git working directory"""
    return repo.repo.git.diff("--name-only").split()


def get_changed_files_committed_and_workdir(
    repo: Git, commithash_to_compare: str
) -> List[str]:
    """Get changed files between given commit and the working copy"""
    return repo.repo.git.diff("--name-only", commithash_to_compare).split()


def get_file_diff_data_workdir(repo: Git, file_path: str) -> str:
    """Get git diff for a file in git working directory"""
    return repo.repo.git.diff("-U0", "--", file_path)


def get_file_diff_data_committed_and_workdir(
    repo: Git, file_path: str, commithash_to_compare: str
) -> str:
    """Get git diff for a file from changes between given commit and the working copy"""
    return repo.repo.git.diff("-U0", commithash_to_compare, "--", file_path)


def get_changed_lines(diff: str) -> Set[int]:
    """Parse changed lines from git diff -U0 output.
    - Change data according to git diff output:
    - @@ -old0,old1 +new0,new1 @@
    - old0 to old0 + old1 are now new0 to new0 + new1
    - changed lines = old0 to old0 + old1 (last not included)
    """
    regex = r"[@][@]\s+[-][0-9]+(?:,[0-9]+)?\s+[+][0-9]+(?:,[0-9]+)?\s+[@][@]"
    line_changes = re.findall(regex, diff)
    changed_lines: Set[int] = set()
    for change in line_changes:
        changed_line = change.strip("@").split()

        # Add , for parsing when it's omitted in the default case of 1
        if "," not in changed_line[0]:
            changed_line[0] += ",1"

        # Split the change to old0 and old1
        old = changed_line[0].split(",")
        old[0] = old[0].strip("-")

        if int(old[1]) == 0:
            changed_lines.add(int(old[0]))
        else:
            changed_lines.update(range(int(old[0]), int(old[0]) + int(old[1])))

    return changed_lines


def get_git_repo() -> Git:
    """Return a Git"""
    res = subprocess.check_output(
        "git rev-parse --show-toplevel".split(),
        stderr=subprocess.DEVNULL,
    )
    project_folder = res.decode().strip()
    return Git(project_folder)


def is_git_repo() -> bool:
    """Check if current directory or any parent directory is a git repo"""
    try:
        get_git_repo()
        return True
    except subprocess.CalledProcessError:
        return False
